-------
-- BroBot
-- Gives the Scientist Scanbot personality
-- by Enos Shenk
-------
-- Adds summon/dismissal random text
-- Adds responses to text from the player beginning with the scanbot's name
-- Adds text said when scanbot scans a location, or when concluding scan
-- Adds an ELIZA module to respond to text in /say from the player if prefixed by the scanbot name
--  or if the bot is already listening from a previous call by his name.
-------
-- v1.0:	Initial release
--
-- v1.1: 	Added an internal cooldown for summon/dismiss talk to prevent spam for Stalkers
--			Fixed a bug where the addon wouldn't work with multiple-word scanbot names
--			Fixed a bug where listen timer wasn't getting reset from base time
--			BroBot will now detect if SAmotes is installed and use strings with emotes.
--

-- Init stuff
require "Unit"
require "GameLib"
require "Window"
require "ApolloTimer"
require "ChatSystemLib"
require "ChatChannelLib"
require "PlayerPathLib"
require "ScientistScanBotProfile"
 
local BroBot = {}

-- Variables
local OurScanBotUnit			-- Current unit reference to our scanbot, if any.
local OurScanBotName			-- Current name of our scanbot. Retained through dismissal.
local OurScanBotNameWords		-- Number of words in our scanbot's name.
local TalkAtScanEnd = false		-- True if text will be displayed at end of scan, as opposed to at scan start. Set per scan call. UNUSED FOR NOW
local IsListening = false		-- If true, bot will interpret /say from owner as input to the ELIZA module
local SaySDText = true			-- If true, bot will speak when summoned or dismissed.
local MessageBuffer = ""		-- Used to hold a message ready for output for time delay

local ListenTimer				-- Timer to reset listening state
local TalkTimer					-- Timer used for queued message sending
local BoredTimer				-- Timer to control when we say a bored string
local SDCooldownTimer			-- Timer used to throttle summon/dismiss text

-- Extras
local UseSAMotes = false
 
function BroBot:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self
    return o
end

function BroBot:Init()
    Apollo.RegisterAddon(self)
end
 
function BroBot:OnLoad()
	Apollo.RegisterSlashCommand("brobot", "OnBrobot", self)
	Apollo.RegisterEventHandler("PlayerPathScientistScanBotDeployed", "OnBrobotDeploy", self)
	Apollo.RegisterEventHandler("PlayerPathScientistScanBotDespawned", "OnBrobotDespawned", self)
	-- These don't fire and I'm sad
	-- Apollo.RegisterEventHandler("PlayerPathScientistScanStart", "OnBrobotScanStart", self)
	--A pollo.RegisterEventHandler("PlayerPathScientistScanEnd", "OnBrobotScanEnd", self)
	Apollo.RegisterEventHandler("PlayerPathScientistScanData", "OnBrobotScanStart", self)
	Apollo.RegisterEventHandler("PetCustomizationUpdated", "OnBrobotCustomize", self)
	Apollo.RegisterEventHandler("ChatMessage", "OnBrobotChatMessage", self)

	ListenTimer = ApolloTimer.Create(20, false, "BrobotListenTimerExpired", self)
	TalkTimer = ApolloTimer.Create(1, false, "BrobotSendMessage", self)
	BoredTimer = ApolloTimer.Create(300, false, "BrobotBored", self)
	SDCooldownTimer = ApolloTimer.Create(5, false, "ThrottleSDEnd", self)
	
	-- See if SAMotes is installed
	if Apollo.GetAddon("SAmotes") ~= nil then
		UseSAMotes = true
	end
end

-- Lots of strings
-- If you want to add your own, make sure you preserve formatting.
-- Entries must be seperated with comma except for the last entry

local BroBotStringsSummon = {
	-- Strings played when bot is summoned
	"Hello!",
	"Greetings.",
	"Hello again, master.",
	"Ready to work!",
	"Systems online.",
	"Another fine day.",
	"It's good to be back."
}

local BroBotStringsSummonSA = {
	"Hello! :D",
	"Greetings.",
	"Hello again, master.",
	"Ready to work! :buddy:",
	"Systems online.",
	"Another fine day.",
	"It's good to be back.",
	":tipshat:",
	":getin:"
}

local BroBotStringsDespawn = {
	-- Strings played when bot is dismissed
	"Goodbye!",
	"Until next time, master.",
	"I know when I'm not wanted.",
	"Powering down.",
	"Goodbye, master.",
	"I shall miss you.",
	"You got it."
}

local BroBotStringsDespawnSA = {
	"Goodbye! :cry:",
	"Until next time, master.",
	":smithicide:",
	":saddowns:",
	"Goodbye, master.",
	"I shall miss you.",
	"You got it.",
	":shrug:"
}

local BroBotStringsAck = {
	-- Strings played when player only says the bots name
	"Yes?",
	"What can I do for you?",
	"Yes master?",
	"Yeah boss?",
	"Hm?",
	"Yeah?",
	"What do you need?"
}

local BroBotStringsAckSA = {
	"Yes?",
	":frogon:",
	":confused:",
	"Yeah boss?",
	"Hm?",
	"Yeah?",
	"What do you need?"
}

local BroBotStringsStartScan = {
	-- Strings played when scan is begun
	"Oooh, how interesting!",
	"Oh, that thing?",
	"It doesn't look that interesting.",	
	"Let me show you what I found!",
	"Fascinating, master.",
	"Oh, I see now!",
	"I found some interesting data.",
	"Data processed and stored.",
	"Scan completed in 2.71 seconds.",
	"Processing data...",
	"A fascinating specimen."
}

local BroBotStringsStartScanSA = {
	":science:",
	":spergin:",
	":vince:",	
	":heysexy:",
	":awesome:",
	":wow:",
	":techno:",
	":tali:",
	":suspense:",
	":jeb:",
	":speculate:"
}

local BroBotStringsBored = {
	-- Strings randomly played when a period of time has gone between interactions
	"Are we just going to stand around all day?",
	"Error: Boredom levels rising, requesting mission.",
	"Master, we haven't scanned anything for ages!",
	"I hope I remember how to scan things when we finally get back to work.",
	"Rust levels rising on my scanner assembly.",
	"Anything to scan? No? Oh well...",
	"Interesting terrain here, master.",
	"So, how is your day master?",
	"You aren't going to replace me, are you master?"
}

local BroBotStringsCustomize = {
	-- Strings played when player customizes scanbot
	"Oh, I feel stylish.",
	"I'm sure the other bots will be jealous.",
	"Thank you, master!",
	"I shall wear it with pride, master.",
	"Thanks!",
	"Error: Critical levels of style reached.",
	"My transistors are positively glowing with pride."
}

-- ELIZA strings
local ElizaRandReplies = {
	"What does that suggest to you?",
	"I see...",
	"I'm not sure I understand you fully.",
	"Can you elaborate on that?",
	"That is quite interesting!",
	"That's so, please continue...",
	"I understand...",
	"Well, well...Do go on.",
	"Why are you saying that?",
	"Please explain the background to that remark...",
	"Could you say that again, in a different way?",
}

local ElizaReplies = {
	[" can you"] = "Perhaps you would like to be able to",
	[" do you"] = "Yes, I",
	[" can i"] = "Perhaps you don't want to be able to",
	[" you are"] = "What makes you think I am",
	[" you're"] = "What is your reaction to me being",
	[" i don't"] = "Why don't you",
	[" i feel"] = "Tell me more about feeling",
	[" why don't you"] = "Why would you want me to",
	[" why can't i"] = "What makes you think you should be able to",
	[" are you"] = "Why are you interested in whether or not I",
	[" i can't"] = "How do you know you can't",
	[" sex"] = "I feel you should discuss this with a human.",
	[" i am"] = "How long have you been",
	[" i'm"] = "Why are you telling me you are",
	[" i want"] = "Why do you want",
	[" what"] = "What do you think?",
	[" how"] = "What answer would please you the most?",
	[" who"] = "How often do you think of such questions?",
	[" where"] = "Why did you think of that?",
	[" when"] = "What would your best friend say to that question?",
	[" why"] = "What is it you really want to know?",
	[" perhaps"] = "You're not very firm on that!",
	[" drink"] = "Moderation in all things should be the rule.",
	[" sorry"] = "Why are you apologizing?",
	[" dreams"] = "Why did you bring up the subject of dreams?",
	[" i like"] = "Is it good that you like",
	[" maybe"] = "Aren't you being a bit tentative?",
	[" no"] = "Why are you being negative?",
	[" your"] = "Why are you concerned about my",
	[" always"] = "Can you think of a specific example?",
	[" think"] = "Do you doubt",
	[" yes"] = "You seem quite certain. Why is this so?",
	[" friend"] = "Why do you bring up the subject of friends?",
	[" computer"] = "Why do you mention computers?",
	[" am i"] = "You are",
}

local ElizaConjugate = {
	[" i "] = "you",
	[" are "] = "am",
	[" were "] = "was",
	[" you "] = "me",
	[" your "] = "my",
	[" i've "] = "you've",
	[" i'm "] = "you're",
	[" me "] = "you",
	[" am i "] = "you are",
	[" am "] = "are",
}

--
-- Code below here
--

-------
-- ELIZA module
-- LUA implementation by Kein-Hong Man (http://lua-users.org/wiki/SciteElizaClassic)
-- Converted to use in this addon by Enos Shenk
-------

function BroBot:Eliza(inText)

	local user = string.lower(inText)
	local response = ""	
	local userOrig = user

	function replyRandomly()
		response = ElizaRandReplies[math.random(#ElizaRandReplies)].."\n"
	end
	
	-- find keyword, phrase
	function processInput()
		for keyword, reply in pairs(ElizaReplies) do
			local d, e = string.find(user, keyword, 1, 1)
			if d then
				-- process keywords
				response = response..reply.." "
				if string.byte(string.sub(reply, -1)) < 65 then -- "A"
					response = response.."\n"; return
				end
				local h = string.len(user) - (d + string.len(keyword))
				if h > 0 then
					user = string.sub(user, -h)
				end
				for cFrom, cTo in pairs(ElizaConjugate) do
					local f, g = string.find(user, cFrom, 1, 1)
					if f then
						local j = string.sub(user, 1, f - 1).." "..cTo
						local z = string.len(user) - (f - 1) - string.len(cTo)
						response = response..j.."\n"
						if z > 2 then
							local l = string.sub(user, -(z - 2))
							if not string.find(userOrig, l) then return end
						end
						if z > 2 then response = response..string.sub(user, -(z - 2)).."\n" end
						if z < 2 then response = response.."\n" end
						return
					end--if f
				end--for
			response = response..user
			return
			end--if d
		end--for
		replyRandomly()
		return
	end
	
	-- main()
	-- accept user input
	if string.sub(user, 1, 3) == "bye" then
		response = "Leaving so soon?"
		return response
	end
	if string.sub(user, 1, 7) == "because" then
		user = string.sub(user, 8)
	end
	user = " "..user.." "
	-- process input, print reply
	processInput()
	
	BroBot:QueueMessage(response)	
end	

-- strsplit code from WildstarNASA
-- Breaks an input string into a table of single words, cut on a provided deliminator.
function BroBot:strsplit(delim, str, maxNb)
    -- Eliminate bad cases...
    if string.find(str, delim) == nil then
        return { str }
    end
    if maxNb == nil or maxNb < 1 then
        maxNb = 0    -- No limit
    end
    local result = {}
    local pat = "(.-)" .. delim .. "()"
    local nb = 0
    local lastPos
    for part, pos in string.gfind(str, pat) do
        nb = nb + 1
        result[nb] = part
        lastPos = pos
        if nb == maxNb then break end
    end
    -- Handle the last field
    if nb ~= maxNb then
        result[nb + 1] = string.sub(str, lastPos)
    end
 
    return result
end

-- strunsplit from me, it's probably bad but it works. 
-- Merges a table of single words back into a string, deliminated by provided character.
function BroBot:strunsplit(delim, inTable)
	local out = ""
	
	for i, v in ipairs(inTable) do
		out = out..v..delim
	end
	
	return out
end

-- Listen timer reset function. Enables listen mode and begins the expire timer fresh. Also refresh the bored timer.
function BroBot:BrobotStartListen()
	IsListening = true
	ListenTimer:Stop()	
	ListenTimer:Start()
	BoredTimer:Stop()
	BoredTimer:Start()
end

-- Listen timer expire function
function BroBot:BrobotListenTimerExpired()
	IsListening = false
end

-- Queue a message into the buffer. Handles setting up the text delay timer.
function BroBot:QueueMessage(inText)
--	Print("Message queued")
	MessageBuffer = inText
	TalkTimer:Set(string.len(inText) / 20, false, "BrobotSendMessage")
	TalkTimer:Start()
end

-- Message sender. Transmits the text to the chat window as well as clearing the buffer.
function BroBot:BrobotSendMessage()
	if MessageBuffer ~= "" then
		ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Say, MessageBuffer, OurScanBotName)
		MessageBuffer = ""
	end
end

-- Punctuation cleaner, removes punctuation marks likely to obscure words like the bot name.
function BroBot:RemovePunctuation(inText)
	inText:gsub("?", "")
	inText:gsub("!", "")
	inText:gsub(",", "")
	inText:gsub(".", "")
	
	return inText
end

-- Called to disable summon/dismiss text until timer expires
function BroBot:ThrottleSD()
	SaySDText = false
	 SDCooldownTimer:Start()
end

-- summon/dismiss timer expired, reset cooldown
function BroBot:ThrottleSDEnd()
	SaySDText = true
end

-- Check the passed in text to see if the first words match the current scanbot name
function BroBot:CheckTextForNameAtStart(inText)
	local nameSplit = BroBot:strsplit(" ", OurScanBotName)
	for i, v in ipairs(nameSplit) do
--		Print("Name: "..v.." Text: "..inText[i])
		if string.lower(inText[i]) ~= string.lower(v) then
			-- Match failure, it's all over.
			return false		
		end 
	end
	-- We made it through without bombing, it's a match
	return true
end

-- Play a random string as we've done nothing interesting for 5 minutes
function BroBot:BrobotBored()
	local HasBot = PlayerPathLib.ScientistHasScanBot()
	if HasBot == true then	
		-- Pick a random string
		local message = BroBotStringsBored[math.random(#BroBotStringsBored)]
		-- Send it
		BroBot:QueueMessage(message)
		BroBot:BrobotStartListen()
	end
end

-- Slash command handler
function BroBot:OnBrobot()
	print("BroBot Online")
end

-- ChatMessage event handler
function BroBot:OnBrobotChatMessage(channelCurrent, tMessage)
	-- Check things. Chat channel, speaker is our owner, bot is deployed. We're only interested in messages that meet these criteria.
	if channelCurrent:GetName() == "Say" then
		if tMessage.strSender == GameLib.GetPlayerUnit():GetName() then
			local HasBot = PlayerPathLib.ScientistHasScanBot()
			if HasBot == true then			
				if IsListening == false then
					-- Bot isn't in listening mode, look for the first word
					local text = BroBot:RemovePunctuation(tMessage.arMessageSegments[1].strText)
					local words = BroBot:strsplit(" ", text)
					if BroBot:CheckTextForNameAtStart(words) == true then
						-- First words of this string match our stored bot name
						if #words == OurScanBotNameWords then
							-- The string is only composed of the bot name, play an acknowledge string and listen for more
							local message = BroBotStringsAck[math.random(#BroBotStringsAck)]
							-- Send it
							--ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Say, message, OurScanBotName)
							BroBot:QueueMessage(message)
							-- Set up listening bool and timer
							BroBot:BrobotStartListen()
						else
							-- String is longer than one word, but our name is the first word. Lop it off and send the rest to ELIZA
						--	table.remove(words, 1)
							local ToEliza = BroBot:strunsplit(" ", words)
							ToEliza = string.sub(ToEliza, string.len(OurScanBotName))
							BroBot:Eliza(ToEliza)
							-- Set up listening bool and timer
							BroBot:BrobotStartListen()
						end						
					end
				else
					-- Bot is listening. If string isn't just our name, send it to ELIZA
					local text = BroBot:RemovePunctuation(tMessage.arMessageSegments[1].strText)
					local words = BroBot:strsplit(" ", text)
					if string.lower(words[1]) == string.lower(OurScanBotName) then
						if #words == 1 then
							-- Only word in the string is our name, play an acknowledge string and listen for more
							local message
							if UseSAMotes == true then
								-- Pick a random string
								message = BroBotStringsAckSA[math.random(#BroBotStringsAckSA)]			
							else
								-- Pick a random string
								message = BroBotStringsAck[math.random(#BroBotStringsAck)]
							end
							-- Send it
							--ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Say, message, OurScanBotName)
							BroBot:QueueMessage(message)
							-- Set up listening bool and timer
							BroBot:BrobotStartListen()
						else
							-- String is longer than one word, but our name is the first word. Lop it off and send the rest to ELIZA
							table.remove(words, 1)
							local ToEliza = BroBot:strunsplit(" ", words)
							BroBot:Eliza(ToEliza)
							-- Set up listening bool and timer
							BroBot:BrobotStartListen()
						end							
					else
						-- First word isn't our name. Send text to ELIZA and reset listen timer
						local ToEliza = BroBot:strunsplit(" ", words)
						BroBot:Eliza(ToEliza)
						-- Set up listening bool and timer
						BroBot:BrobotStartListen()
					end
				end				
			end
		end
	end
end

-- Bot deploy handler
function BroBot:OnBrobotDeploy()
	-- Update the bot unit and name variables
	OurScanBotUnit = PlayerPathLib:ScientistGetScanBotUnit()
	OurScanBotName = OurScanBotUnit:GetName()
	local NameWords = BroBot:strsplit(" ", OurScanBotName)
	OurScanBotNameWords = #NameWords
--	Print("Set name: "..OurScanBotName.." Word count: "..OurScanBotNameWords)
	
	if SaySDText == true then
		local message
		if UseSAMotes == true then
			-- Pick a random string
			message = BroBotStringsSummonSA[math.random(#BroBotStringsSummonSA)]			
		else
			-- Pick a random string
			message = BroBotStringsSummon[math.random(#BroBotStringsSummon)]
		end
		-- Send it
		--ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Say, message, OurScanBotName)
		BroBot:QueueMessage(message)
		-- Enable throttle cooldown timer
		BroBot:ThrottleSD()
	end
end

-- Bot despawn handler
function BroBot:OnBrobotDespawned()
	-- Nil out the scanbot unit variable, keep the last seen name.
	OurScanBotUnit = nil
	
	if SaySDText == true then
		local message
		if UseSAMotes == true then
			-- Pick a random string
			message = BroBotStringsDespawnSA[math.random(#BroBotStringsDespawnSA)]			
		else
			-- Pick a random string
			message = BroBotStringsDespawn[math.random(#BroBotStringsDespawn)]
		end
		-- Send it immediately with no queue
		ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Say, message, OurScanBotName)	
		-- Enable throttle cooldown timer
		BroBot:ThrottleSD()
	end
end

-- Scan begin handler
function BroBot:OnBrobotScanStart()
	local message
	if UseSAMotes == true then
		-- Pick a random string
		message = BroBotStringsStartScanSA[math.random(#BroBotStringsStartScanSA)]			
	else
		-- Pick a random string
		message = BroBotStringsStartScan[math.random(#BroBotStringsStartScan)]
	end
	-- Send it
	BroBot:QueueMessage(message)
	BroBot:BrobotStartListen()
end

-- Pet customize updated handler
function BroBot:OnBrobotCustomize()
	local HasBot = PlayerPathLib.ScientistHasScanBot()
	if HasBot == true then	
		-- Update the bot name variables in case they were changed
		OurScanBotName = OurScanBotUnit:GetName()
		local NameWords = BroBot:strsplit(" ", OurScanBotName)
		OurScanBotNameWords = #NameWords
		-- Pick a random string
		local message = BroBotStringsCustomize[math.random(#BroBotStringsCustomize)]
		-- Send it
		BroBot:QueueMessage(message)
		BroBot:BrobotStartListen()
	end
end


-- Make stuff do stuff
local BroBotInst = BroBot:new()
BroBotInst:Init()